module AE



class TextureResizer



class Bounds
# TODO: if no entities given, use convex hull
  ##
  # Find the best fitting BoundingBox (some simplifications for speed)
  # strategy: - collect all vectors in selection (theoretically: all possible vectors between all points)
  #           - create all possible transformations from 2 vectors, transform all points
  #           - get SketchUp's BoundingBox (which is aligned to model axes)
  #           - test against criteria (minimal sum of edge lengths) and take the best BoundingBox
  def initialize(*args)
    return if args.flatten.empty?
    dof = (args.last.class==Fixnum)? args.pop : 3 # optional last parameter: degrees of freedom
    ents = args.flatten if args.first.is_a?(Array)
    if ents.length == 1 && ents[0].class == Sketchup::Face || dof == 2 # 2d
      (face = true; n = ents[0].normal)
    end
    points, vecs = [], []
    # get all points
    ents.each_with_index{|e, i| points << e if e.is_a?(Geom::Point3d)
      points.concat(e.vertices.collect{|v| v.position}) if e.respond_to?("vertices")
    }
    points.uniq!
    # 3 degrees of freedom to optimize
    if dof == 3
      # get vectors; simplification: only vectors of adjacent points
      (points.length-2).times{|i| v = points[i].vector_to(points[i+1]); vecs << v if v.valid?}
      # simplification: consider only the 10 longest edges
      vecs = vecs.sort{|v1,v2| v2.length <=> v1.length}[0...10]
      @t = Geom::Transformation.new
      b = Geom::BoundingBox.new
      b.add(*points)
      @b_max = b.max
      @b_min = b.min
      u = b.width + b.depth + b.height
      # Empiric: mix in some additional vectors between the 10 most distant points for better accuracy
      c = b.center
      l = 10 # max amount of test vectors.
      points.sort!{|p1,p2| p2.distance(c) <=> p1.distance(c)}
      [l, points.length].min.times{|i|
        points[0,l].each{|p|
          v = p.vector_to(points[i]).normalize
          vecs << v if v.valid?
        }
      }
      vecs.uniq!
      vecs.each{|x|
        if face # 2d
          y = x * n
          # flip vectors if necessary to get a reasonable orientation
          x,y = y,x if x.z.abs > y.z.abs
          y.reverse! if y.z < 0
          x.reverse! if !(x*y).samedirection?(n)
          t1 = Geom::Transformation.new(points[0], x, y)
          t1i = t1.inverse
          points.each{|p| p.transform!(t1i)}
          # Update SketchUp's BoundingBox
          b.clear; b.add(*points)
          # Check if it this BoundingBox is better (~ has less circumfence or shorter diagonal)
          if b.width + b.height < u
            u = b.width + b.height
            @t = t1
            @b_max = b.max
            @b_min = b.min
          end
          points.each{|p| p.transform!(t1)}
        else # 3d
          vecs.each{|n|
            next if n.parallel?(x)
            y = (x * n).normalize
            z = x * y
            x, y, z = *([x, y, z].sort{|i,j| i.z.abs <=> j.z.abs})
            y.reverse! if y.z < 0
            x.reverse! if (x*y).z < 0
            t1 = Geom::Transformation.new(points[0], x, y)
            t1i = t1.inverse
            points.each{|p| p.transform!(t1i)}
            # Update SketchUp's BoundingBox
            b.clear; b.add(*points)
            # Check if it this BoundingBox is better (~ shorter diagonal)
            if b.width + b.depth + b.height < u
              u = b.width + b.depth + b.height
              @t = t1
              @b_max = b.max
              @b_min = b.min
            end
            points.each{|p| p.transform!(t1)}
          }
        end
      }
    # 2 degrees of freedom to optimize => simpler BoundingBox with heading and tilt, but roll=0
    elsif dof == 2
      y_vec = Geom::Vector3d.new( ((n!=[0,0,1])?[0,0,1]:[0,1,0]).project_to_plane([0,0,0], n) ).normalize
      x_vec = y_vec * n
      @t = Geom::Transformation.new(x_vec, y_vec, n, points[0])
      ti = @t.inverse
      allx, ally, allz = [], [], []
      points.each{|point|
        p = point.transform(ti)
        allx << p.x
        ally << p.y
        allz << p.z
      }
      @b_min = Geom::Point3d.new(allx.min, ally.min, allz.min)
      @b_max = Geom::Point3d.new(allx.max, ally.max, allz.max)
    end # if dof == 2
  end

  def max
    return @b_max.transform!(@t)
  end

  def min
    return @b_min.transform!(@t)
  end

  def width
    return (@b_max.x - @b_min.x).abs
  end

  def depth
    return (@b_max.y - @b_min.y).abs
  end

  def height
    return (@b_max.z - @b_min.z).abs
  end

  def diagonal
    return Math.sqrt( self.width**2 + self.depth**2 + self.height**2 )
  end

  def corner(number)
    p = number.to_s(2).rjust(3).split(//).collect{|s| s.to_i}.reverse
    x = @b_min.x + p.x * (@b_max.x - @b_min.x)
    y = @b_min.y + p.y * (@b_max.y - @b_min.y)
    z = @b_min.z + p.z * (@b_max.z - @b_min.z)
    return Geom::Point3d.new(x,y,z).transform!(@t)
  end

  def center
    return Geom::Point3d.linear_combination(0.5, @b_min, 0.5, @b_max).transform!(@t)
  end

  def draw
    # Sketchup.active_model.start_operation "Draw Bounds"
    g = Sketchup.active_model.active_entities.add_group
    ps = []
    8.times{|i| ps << self.corner(i)}
    {0=>1,2=>3,4=>5,6=>7}.each_pair{|i,j| l=g.entities.add_line(ps[i],ps[j]);l.material="red" if l}
    {0=>2,1=>3,4=>6,5=>7}.each_pair{|i,j| l=g.entities.add_line(ps[i],ps[j]);l.material="green" if l}
    {0=>4,1=>5,2=>6,3=>7}.each_pair{|i,j| l=g.entities.add_line(ps[i],ps[j]);l.material="blue" if l}
    g.entities.add_cpoint(self.min)
    # Sketchup.active_model.commit_operation
  end

  def transformation
    return @t
  end

  private

  def convex_hull
# TODO
  end

end # class Bounds



end # class TextureResizer



end # module AE
