=begin

Copyright 2011-2013, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         TextureResizer.rb
Author:       Andreas Eisenbarth
Description:  This Plugin allows to downsample all textures in the model by
              percentage or depending on texture size in model.
              This version now also considers the texture's dimensions on individual
              faces (whereas the previous version took the dimensions as set in
              the materials browser) and in transformed/scaled and nested components.
              If two faces have different texture dimensions, the dimensions will
              be averaged (weighted by face area).
              Textures can also be optimized for a specific view if the plugin Goldilocks is installed.
Usage:        menu Plugins > 'Resize Textures'
Requirements: ImageMagick
                Windows: download it from this plugin's site or from the official site
                  http://www.imagemagick.org/script/binary-releases.php#windows
                  Place it either in the Plugins folder or at the location of your choice.
                  The Plugin will then ask you once for the install location.
                Linux (Wine): Your distribution most likely comes with native ImageMagick preinstalled,
                  if not install it from the repositories
                  or install the Windows version of ImageMagick into Wine as described above.
Version:      1.5.7
Date:         27.05.2013

=end

require('sketchup.rb')



module AE



class TextureResizer



# This plugin's folder.
PATH_ROOT = File.dirname(__FILE__) unless defined?(self::PATH_ROOT)
# ImageMagick library.
require(File.join(PATH_ROOT, 'ImageMagick.rb'))
# Translation library.
require(File.join(PATH_ROOT, 'Translate.rb'))
# Load translation strings and add plugin to UI.
TRANSLATE = Translate.new("TextureResizer", File.join(PATH_ROOT, "lang")) unless defined?(self::TRANSLATE)
# WebDialog Helper.
require(File.join(PATH_ROOT, 'Dialog.rb'))
# Options.
require(File.join(PATH_ROOT, 'Options.rb'))
@@options ||= Options.new("TextureResizer", {
  # Method of resizing; :percent or depending on real-world :resolution.
  :method => :percent,
  :percentage => 50,
  :resolution => 20,
  # Whether to consider the uvq coordinates of each individual face (slow) or the dimensions in material browser (fast).
  :resolution_by_faces => true,
  # Whether to limit the possible texture dimensions to a maximum or minimum.
  :limit => false,
  :max => 1024,
  :min => 20,
  # Whether to increase texture sizes if method is "resolution".
  :only_downsample => true,
  # Percentage that the new texture dimensions must differ from the old dimensions.
  :not_modify_threshold => 2,
  # Whether to apply a specific level of jpeg compression, otherwise take it from the source file.
  :compression => false,
  :jpeg_quality => 75,
  :crop => false,
  # Whether to rotate the crop bounding box so that it fits best.
  :crop_rotate => false,
  # resolution to use if the plugin Goldilocks is installed
  :goldilocks_resolution => 4,
  # If the selection specifies the faces/entities, only on selected faces will materials be affected (creates copies if necessary).
  # If the selection specifies the materials, those materials will be resized in the whole model (:global).
  :selection_mode => :faces,
  # If something is selected: Whether materials will be affected only within the selection or globally.
  # If materials are also used outside of the selection, they will copied.
  :selected => false,
})
# Class to get the optimal bounding box.
require(File.join(PATH_ROOT, 'Bounds.rb'))



# Load translation strings and add plugin to UI.
#
unless file_loaded?(__FILE__)
  cmd_resize = UI::Command.new(TRANSLATE["Resize Textures"]) { TextureResizer.run() }
  cmd_resize.tooltip = TRANSLATE["Downsamples all textures in the model by percentage or dependent on texture size in model."]
  cmd_resize.small_icon = File.join(PATH_ROOT, "images", "icon_textureresizer_16.png")
  cmd_resize.large_icon = File.join(PATH_ROOT, "images", "icon_textureresizer_24.png")
  UI.menu("Plugins").add_item(cmd_resize)
  file_loaded(__FILE__)
end # unless file_loaded



@@first_run = true # if the plugin runs for the first time in this session
@@tw = nil
@@optionsdlg = nil # This will hold the options dialog.



# Run plugin from UI and check requirements.
#
def self.run
  if @@first_run == true
    @@tw = Sketchup.create_texture_writer
    @@first_run = false
  end
  # Check if Goldilocks is installed
  @@goldilocks = !!defined?(Goldilocks) && Goldilocks.respond_to?("methods")
  return UI.messagebox(TRANSLATE["The model does not contain any materials."], MB_OK) if Sketchup.active_model.materials.length == 0
  return puts("ImageMagick not installed") unless ImageMagick.installed?
  self.options_dialog
end # def run



# Open the options dialog.
#
def self.options_dialog
  optdlg_title = (Sketchup.active_model.selection.empty?) ? TRANSLATE["Resize textures in model"] : TRANSLATE["Resize textures in selection"]
  @@optionsdlg = AE::TextureResizer::Dialog.new(optdlg_title, false, "AE_TextureResizer", 450, 350, 300, 300, true)
  @@optionsdlg.min_width = 200
  @@optionsdlg.min_height = 200
  html_path = File.join(PATH_ROOT, "html", "TextureResizerOptions.html")
  @@optionsdlg.set_file(html_path)

  # Callbacks

  # init: Pass the default options to the form.
  @@optionsdlg.on_show { |msg|
    TRANSLATE.webdialog(@@optionsdlg)
    @@optionsdlg.execute_script("function() { var s = AE.$('#style0'); s.parentNode.removeChild(s); }()") if @@goldilocks == true
    @@optionsdlg.execute_script("AE.Style.show( AE.$('#selection') );") unless Sketchup.active_model.selection.empty?
    @@optionsdlg.call_function("AE.TextureResizer.initialize", @@options.get_all)
    @@optionsdlg.set_position_center
  }

  # Apply button
  @@optionsdlg.on("update_options") { |msg, hash|
    @@options.update(hash)
    @@options.save
    @@optionsdlg.call_function("AE.TextureResizer.Progress.start")
    # Allow the webdialog to update/redraw.
    t = UI.start_timer(0.1, false){
      self.process() # Takes default options.
    }
  }

  # Release the reference to the dialog, so it isn't cached.
  @@optionsdlg.on_close{
    @@optionsdlg = nil
  }

  # for debugging
  @@optionsdlg.on("puts") { |msg, param|
    puts(param)
  }

  # Show the webdialog.
  @@optionsdlg.show_modal

  return @@optionsdlg
end # def options_dialog



# Main process. Can be invoked directly from ruby without UI.
# @param [Hash] options  An optional hash of options (missing options will be taken from default options)
# @param [Array, Sketchup::EntityCollection, Sketchup::Face] entities  If not given, selected entities or whole model will be used.
# @return nil
#
def self.process(options={},
                 entities=(Sketchup.active_model.selection.empty?) ?
                   Sketchup.active_model.entities :
                   Sketchup.active_model.selection.to_a)
  entities = [entities] if entities.is_a?(Sketchup::Entity) && !entities.is_a?(Sketchup::EntityCollection)
  @@goldilocks = !!defined?(Goldilocks) && Goldilocks.respond_to?("methods")
  options[:selected] = !entities[0].model.selection.empty?
  self.new(entities, options)
end


# Instance methods for processing (using a class instance avoids clashes if several models would be processed simultaneously).
# * Loops over all entities and collects materials from entities.
# * Loops over collected materials and loads them into ImageMagick class.
# * Executes all convert commands.
# @param [Hash] options A hash of options (missing options will be taken from default options)
# @param [Array, Sketchup::EntityCollection] entities
# @return nil
#
def initialize(entities, options={})
  raise(ArgumentError, "No entities given. #{entities.inspect}") unless (entities.is_a?(Array) && !entities.empty?) || entities.is_a?(Sketchup::Entities)
  @model = entities[0].model
  @entities = entities
  @view = @model.active_view
  @exited = false
  @success = nil

  begin
    (Sketchup.version_number < 7000000) ? @model.start_operation(TRANSLATE["Resize Textures"]) : @model.start_operation(TRANSLATE["Resize Textures"], true)
    @options = @@options.get_all.merge(options)
    @materials = {}
    @faces = {}
    @instances = {} # remember each processed definition and of one instance transformation.to_a

    # (shortcut for all materials in model & percentage)
    if @entities.is_a?(Sketchup::Entities) && @entities.parent == @model && !@options[:crop] && @options[:method] != :resolution
      @model.materials.find_all{ |e| e.materialType > 0}.each{ |e| @materials[e] = {} }
    # loop over all entities and collect materials in @materials hash (normal)
    else
      # Collect all faces.
      @faces.clear
      block = Proc.new{ |ent, scale|
        if ent.is_a?(Sketchup::Face)
          @faces[ent] = scale
        elsif ent.is_a?(Sketchup::Group) || ent.is_a?(Sketchup::ComponentInstance)
          definition = (ent.is_a?(Sketchup::Group)) ? ent.entities.parent : ent.definition
          t = ent.transformation
          # Shortcut: If instances have all the same transformation
          # (ie. if this one's transformation equals the first instance's transformation),
          # process only one instance of each definition.
          break if @instances.include?(definition) && t == @instances[definition]
          @instances[definition] ||= t
          t = t.to_a
          # Average scale of the instance.
          s = (Geom::Vector3d.new(t[0],t[4],t[8]).length + Geom::Vector3d.new(t[1],t[5],t[9]).length + Geom::Vector3d.new(t[2],t[6],t[10]).length)/3.0
          definition.entities.each{ |e| block.call(e, s*scale) }
        end
      }

      # Get all faces (in selection) to analyze and their scale (transformation).
      @entities.each{ |e| block.call(e, 1) }

      # Get all materials (in selection) to analyze.
      @faces.each{ |face, scale|
        [face.material, face.back_material].each{ |m|
          @materials[m] ||= {} unless m == nil || m.materialType == 0
        }
      }
      if @options[:selected] && @options[:selection_mode] == :global
        # In this case, all selected materials should be analyzed in whole model.
        @faces.clear
        @model.entities.each{ |e| block.call(e, 1) }
      end

      # Analyze the faces.
      @faces.each{ |face, scale|
        analyze_face(face, scale)
      }
    end

    # Error if materials have no textures.
    if @materials.empty?
      messagebox_text = (options[:selected]) ? TRANSLATE["The selection does not contain any textures."] : TRANSLATE["The model does not contain any textures."]
      @@optionsdlg.close unless @@optionsdlg.nil?
      @model.abort_operation
      UI.messagebox(messagebox_text, MB_OK)
    end

    # Analyze model with Goldilocks.
    if @options[:method] == :goldilocks
      return puts(TRANSLATE["Option 'goldilocks' requested but goldilocks.rb is not installed."]) if @@goldilocks == false
      goldilocks_analyze(@entities)
    end # if @@goldilocks

    # Load each material into ImageMagick and create convert commands.
    @imagemagick = ImageMagick.caches[@model] || ImageMagick.new
    @materials.each_key{ |m|
      prepare_conversion(m)
    }

    # Work off the queue.
    @imagemagick.execute{
      restore_cropped if @options[:crop] # Before update_materials because it has to look up face.material in @materials!
      update_materials if @options[:selected] && @options[:selection_mode] == :faces # Important!: Breaks also correlation between face.material and @materials[material].
      @@optionsdlg.call_function("AE.TextureResizer.Progress.end") unless @@optionsdlg.nil?
      @@optionsdlg.close if !@@optionsdlg.nil? && @@optionsdlg.visible?
      @imagemagick.purge # TODO: optional. Could also leave materials in cache (but then undo observer required).
      @model.commit_operation
      if @options[:method] == :goldilocks && @options[:open_goldilocks] == true
       (Goldilocks.respond_to?(:texreport))? Goldilocks.texreport : Goldilocks.report
      end
      @success = true
      @exited = true
    }

    return nil

  rescue Exception => e
    # Clean up temporary files.
    @imagemagick.erase unless @imagemagick.nil?
    @@optionsdlg.close unless @@optionsdlg.nil?
    @model.abort_operation
    puts("Error #<#{e.class.name}:#{e.message}.>")
    puts(e.backtrace)
    @success = false
    @exited = true
  end # begin/rescue
end # def initialize



# Tells whether the whole resizing process has exited.
# @return [Boolean]
#
attr_reader :exited
alias :exited? :exited


# Tells whether the whole resizing process was successfully.
# @return [Boolean]
#
attr_reader :success
alias :success? :success


# Tells whether the whole resizing process has exited successfully.
# @return [Boolean]
#
def completed
  return @exited && @success
end
alias :completed? :completed



# Analyze a Face for texture resolution and area, add the result to @materials hash.
# @param [Sketchup::Face] e  An entity to analyze.
# @param [Float] scale  optional scale of the parent's transformation (parent Group/Component)
#
def analyze_face(face, scale=1)
  # For both front and back material do
  [face.material, face.back_material].each_with_index{ |m, i|
    next unless @materials.include?(m)
    # This material should be resized.
    # => It is in the hash.
    side = (i == 0) ? true : false
    if @options[:crop]
      # This material should be cropped.
      # => Get the uv coordinates, to see how much of the image does not cover faces.
      uvh = face.get_UVHelper(true, true, @@tw)
      # Translate all UVs so that the center of the face is in the range 0-1.
      # This is necessary if faces share a repeated tile of the same area on the texture.
      cuv = (side) ? uvh.get_front_UVQ(face.bounds.center) : uvh.get_back_UVQ(face.bounds.center)
      cuv.x/=cuv.z; cuv.y/=cuv.z; cuv.z=1
      translation = Geom::Vector3d.new(-cuv.x.divmod(1)[0], -cuv.y.divmod(1)[0], 0)
      # Get all vertices' UVs.
      uvs = face.vertices.collect{ |v|
        uv = (side) ? uvh.get_front_UVQ(v.position) : uvh.get_back_UVQ(v.position)
        uv.x/=uv.z; uv.y/=uv.z; uv.z=1
        uv += translation
        uv
      }
      @materials[m][:uv] ||= []
      @materials[m][:uv].push(*uvs)
    end
    if @options[:method] == :resolution && @options[:resolution_by_faces] == true
      # This material should be resized by texture resolution.
      # => We need the resolution and the area (as weight).
      area = face.area
      res = ImageMagick.texture_resolution(face, side) # one texture unit in model units
      # Get the resolution in x and y direction.
      @materials[m][:res] ||= 0
      @materials[m][:res] += res * scale * area
      # Get the area.
      @materials[m][:area] ||= 0
      @materials[m][:area] += area
    end
  }
end # def analyze_face
private :analyze_face



# Analyze model with Goldilocks.
#   Following lines based on Goldilocks, Copyright (c) 2009 Billyard Enterprises.
#
def goldilocks_analyze(entities)
  width = @view.vpwidth
  height = @view.vpheight
  unitvp = Math::sqrt(width * height)
  stats = Hash.new
  w2t = Hash.new
  @materials.each { |mat, value|
    w2t[mat] = Math::sqrt(mat.texture.image_width * mat.texture.image_height) / Math::sqrt(mat.texture.width * mat.texture.height) if mat.texture
  }
  Goldilocks.walkFaces(entities, Geom::Transformation.new, nil, stats) { |face, xform, zdist, mat|
    tsize = Math::sqrt(face.area) * w2t[mat]
    psize = Math::sqrt(face.area) * (unitvp / zdist)
    tsize / psize
  }
  @materials.each{ |mat, value| value[:ratio] = stats[mat] }
end # def goldilocks_analyze
private :goldilocks_analyze



# Prepare a material for conversion. Apply the options to create a convert command.
# @param [Sketchup::Material] material
#
def prepare_conversion(material)
  # Export texture path.
  texturepath = @imagemagick.load(material, false)
  if texturepath.nil?
    puts(TRANSLATE["skip %0: could not load into cache", material.name])
    return @materials.delete(material)
  end

  pw = material.texture.image_width # original pixel width
  ph = material.texture.image_height # original pixel height
  rw = material.texture.width.to_f # real world width
  rh = material.texture.height.to_f # real world height
  percentage = @options[:percentage].to_f
  target_resolution = @options[:resolution].to_f / 1.m
  max = @options[:max].to_f
  min = @options[:min].to_f
  convert_option = ""
  nw  = pw # new pixel width
  nh = ph # new pixel height

  # Crop
  if @options[:crop] && @materials[material].include?(:uv)
    # First try optimization of the BoundingBox. This won't allow tiling, so we could get a bigger image size than wanted.
    # Little detour: Convert UV coordinates into pixel coordinates because the image could be rectangular.
    uv2px = Geom::Transformation.scaling(pw, ph, 1)
    px = @materials[material][:uv].collect{ |uv| uv.transform(uv2px)}
    b = AE::TextureResizer::Bounds.new(px) # in pixel space
    corners = [b.corner(2), b.corner(0), b.corner(1), b.corner(3)] # counter-clockwise from top left
    new_area = b.depth * b.width
    # Check that optimized BoundingBox results in a smaller image.
    if new_area < pw*ph && @options[:crop_rotate]
      # Transformation
      axis = Geom::Transformation.new(corners[1], corners[1].vector_to(corners[2]), corners[2].vector_to(corners[3]))
      scaling = Geom::Transformation.scaling(corners[1].distance(corners[2]), corners[0].distance(corners[1]), 1)
      @materials[material][:crop_transform] = uv2px.inverse * axis * scaling
      # @materials[material][:crop_size] = [b.width, b.height] # deprecated
      @materials[material][:cropped] = true
      # create command for ImageMagick
      xy = []
      corners.each{ |uv| xy << (uv[0]).to_i.to_s+","+(ph-(uv[1])).to_i.to_s}
      convert_option += %[-virtual-pixel tile -distort Perspective ]
      convert_option +=  %["#{xy[0]} 0,0  #{xy[1]} 0,#{ph} ]
      convert_option += %[ #{xy[2]} #{pw},#{ph}  #{xy[3]} #{pw},0"]
      # update the pixel dimensions
      nw = b.width.to_f# * pw
      nh = b.depth.to_f# * ph
    # Otherwise do simple rectangular cropping (works with tiling). Here we can use UV coordinates also for rectangular images.
    else
      b = Geom::BoundingBox.new
      b.add(*@materials[material][:uv]) # in UV space
      # Don't crop bigger than the original image => intersect the BoundingBox with a 1×1 BoundingBox
      b = b.intersect(Geom::BoundingBox.new.add(ORIGIN, [1,1,1]))
      # BoundingBox.corners: red, green, blue, yellow
      corners = [b.corner(2), b.corner(0), b.corner(1), b.corner(3)] # counter-clockwise from top left
      # Transformation
      translation = Geom::Transformation.new(Geom::Vector3d.new(corners[0].to_a))
      scaling = Geom::Transformation.scaling(corners[1].distance(corners[2]), corners[0].distance(corners[1]), 1)
      @materials[material][:crop_transform] = translation * scaling
      # @materials[material][:crop_size] = [b.width, b.height] # deprecated
      @materials[material][:cropped] = true
      # update the pixel dimensions
      nw = b.width.to_f * pw
      nh = b.height.to_f * ph
      # create command for ImageMagick
      cl = corners[0].x * pw
      ct = ph - corners[0].y * ph
      convert_option += " -crop #{nw.to_i}x#{nh.to_i}+#{cl.to_i}+#{ct.to_i}"
    end
  end

  # Calculate new dimensions.
  if @options[:method] == :percent
    nw *= percentage / 100.0 # new pixel width
    nh *= percentage / 100.0 # new pixel height
  elsif @options[:method] == :resolution
    if @materials && @materials[material] && @materials[material][:res]
      current_resolution = @materials[material][:res] / @materials[material][:area]
      nw *= target_resolution / current_resolution
      nh *= target_resolution / current_resolution
    else
      nw = rw * target_resolution
      nh = rh * target_resolution
    end
  elsif @options[:method] == :goldilocks
    ratio = @options[:goldilocks_resolution].to_f / @materials[material][:ratio].to_f
    return @materials.delete(material) if ratio > 1
    nw *= ratio
    nh *= ratio
  end

  # Set the limit.
  if @options[:limit] == true
    (nh = nh/nw*min; nw = min) if nw < min
    (nw = nw/nh*min; nh = min) if nh < min
    (nh = nh/nw*max; nw = max) if nw > max
    (nw = nw/nh*max; nh = max) if nh > max
    nw = min if nw < min # if one side of texture is below min limit after adjusting it to max limit
    nh = min if nh < min # force it into both limits
  end

  # Avoid zero pixel dimensions or ugly color artefacts at too small sizes.
  nw = 4 if nw < 4
  nh = 4 if nh < 4

  # Optionally skip texture if new dimensions would be bigger than old dimensions.
  if @options[:only_downsample] == true && @options[:method] == :resolution && (nw > pw+1 || nh > ph+1)
    @materials.delete(material)
    puts(TRANSLATE["skip %0: new dimensions %1 would be bigger than old dimensions %2", material.name, "#{nw}x#{nh}", "#{pw}x#{ph}"])
    return
  end

  # Skip the texture if the new texture dimensions differ less than <threshold> % from the old dimensions.
  if (100*(1-nw/pw.to_f).abs < @options[:not_modify_threshold] ||
      100*(1-nh/ph.to_f).abs < @options[:not_modify_threshold]) && @options[:compression] != true && !@materials[material][:cropped]
    @materials.delete(material)
    puts(TRANSLATE["skip %0: new dimensions almost identical to old dimensions", material.name])
    return
  end
  # Compose the command for ImageMagick.
  # example: convert.exe input.jpg -resize 256x512! output.jpg
  # convert_option: -resize <width>x<height>{%}{@}{!}{<}{>}
  convert_option += " -resize #{nw.to_i}x#{nh.to_i}!"
  # Set the quality for jpg or jpeg textures.
  convert_option += " -quality " + @options[:jpeg_quality].to_s if @options[:compression] == true && texturepath[/\.(?:jpg|jpeg)$/] != nil

  # Add the command to the queue:
  # Create a new material (copies) to preserve rest of model.
  if @options[:selected] && @options[:selection_mode] == :faces
    @imagemagick.create(material, convert_option, true){ |new_material| # true: put command into queue until "execute" is explicitely called
      # Update the materials hash.
      @materials[material][:copy] = new_material
    }
  # Or modify the original material.
  else
    @imagemagick.edit(material, convert_option, true) # true: adds the command into queue until "execute" is explicitely called
  end

end # def prepare_conversion
private :prepare_conversion



# Replace a material on a face by a copy (if a copy has been created).
# This is useful if only some but not all faces of a material should be modified.
# The original materials stays untouched and a copy is created.
# Important: This method breaks the correlation between a face's material and the metadata in @materials.
#
def update_materials
  @faces.each{ |face, scale|
    # For both front and back material do:
    hash = @materials[face.material]
    face.material = hash[:copy] if hash && hash[:copy]
    hash = @materials[face.back_material]
    face.back_material = hash[:copy] if hash && hash[:copy]
  }
end # def update_materials
private :update_materials



# This method updates the UV coordinates for materials that have been cropped.
#
def restore_cropped
  @faces.each{ |face, scale|
    # For both front and back material do:
    [face.material, face.back_material].each_with_index{ |material,i|
      next unless @materials.include?(material)
      next unless @materials[material][:cropped]
      side = (i == 0) ? true : false
      uvh = face.get_UVHelper(true, true, @@tw)
      # Translate all UVs so that the center of the face is in the range 0-1.
      # This is necessary if faces share a repeated tile of the same area on the texture.
      cuv = (side) ? uvh.get_front_UVQ(face.bounds.center) : uvh.get_back_UVQ(face.bounds.center)
      cuv.x/=cuv.z; cuv.y/=cuv.z; cuv.z=1
      translation = Geom::Vector3d.new(-cuv.x.divmod(1)[0], -cuv.y.divmod(1)[0], 0)
      # Get all vertices' UVs.
      ps = face.vertices.collect{ |v| v.position}
      ps << ps[0] + ps[1].vector_to(ps[2]) if ps.length == 3 # triangle
      uvs = ps.collect{ |p|
        uv = (side) ? uvh.get_front_UVQ(p) : uvh.get_back_UVQ(p)
        uv.x/=uv.z; uv.y/=uv.z; uv.z=1
        uv += translation
        uv
      }
      # Transform UVs to coordinate system of crop bounding box.
      t = @materials[material][:crop_transform].inverse
      uvqs = uvs.collect{ |uv| uv.transform(t) }
      # Swap the material against a copy.
      material = @materials[material][:copy] if @materials[material][:copy]
      # Position texture on all faces.
      pt_array = []
      4.times{ |i| pt_array << ps[i]; pt_array << uvqs[i]}
      face.position_material(material, pt_array, side) rescue nil # Could not compute valid matrix from points
    }
  }
end # def restore_cropped
private :restore_cropped



end # class TextureResizer



end # module AE
